import datetime
import logging
import os
import sys
import signal
import time
import uuid

import requests
from enum import Enum
from croniter import croniter
from datetime import timedelta

"""
Important ENV_VARS:

API_KEY - authenticates with api server
TARGET_HOST - sets the canary target
CANARY_ID - sets the canary id to uniquely associate data
TEMPLATE_LIBRARY_ID - sets the canary template library id to uniquely generate health, compliance and forecast data
CANARY_API_SERVER - location of reporting server
"""

"""
Extra ENV_VARS:

IS_AGENT_CANARY - sets the canary to be Agent Canary that can run outside Kubernetes cluster or airflow
CANARY_SCHEDULE - defines the schedule that the Canary Agent will run based on
"""


CANARY_TYPE_NAME = os.getenv("CANARY_TYPE")
TEMPLATE_LIBRARY_ID = os.getenv("TEMPLATE_LIBRARY_ID")

try:
    IS_AGENT_CANARY = bool(os.getenv("IS_AGENT_CANARY", False))
except:
    IS_AGENT_CANARY = False

try:
    CANARY_SCHEDULE = os.getenv("CANARY_SCHEDULE", "")
except:
    CANARY_SCHEDULE = ""

logging.basicConfig(
    level=logging.INFO,
    format="%(levelname)s %(asctime)s %(name)s.%(funcName)s:%(lineno)s- %(message)s",
)

logger = logging.getLogger(__name__)


class CanaryRunStatus(Enum):
    RUNNING = "RUNNING"
    STOPPED = "STOPPED"
    ERROR = "ERROR"
    NEW = "NEW"


class GracefulKiller:
    kill_now = False

    def __init__(self):
        signal.signal(signal.SIGINT, self.exit_gracefully)
        signal.signal(signal.SIGTERM, self.exit_gracefully)

    def exit_gracefully(self, *args):
        self.kill_now = True


def roundDownTime(dt=None, dateDelta=timedelta(minutes=1)):
    roundTo = dateDelta.total_seconds()
    if dt == None:
        dt = datetime.datetime.now()
    seconds = (dt - dt.min).seconds
    rounding = (seconds + roundTo / 2) // roundTo * roundTo
    return dt + timedelta(0, rounding - seconds, -dt.microsecond)


def getNextCronRunTime(schedule):
    return croniter(schedule, datetime.datetime.now()).get_next(datetime.datetime)


def sleepTillTopOfNextMinute():
    t = datetime.datetime.utcnow()
    sleeptime = 60 - (t.second + t.microsecond / 1000000.0)
    time.sleep(sleeptime)


try:
    sys.path.append("/CC")
except Exception as e:
    logger.error("Error adding path %s", e)

try:
    sys.path.append(
        os.path.join(
            "./",
        )
    )
except Exception as e:
    logger.error("Error adding path %s", e)


"""
    Common methods for both canaries here and how to interact with the Cloud Canaries System
"""


class CanaryBasePrototype(object):
    """The base class for a Canary"""

    target_host = None
    target_value = None
    first_run = True

    def __init__(self):
        self.startup_logger = logging.getLogger(__name__)
        self._error_get = False
        self._error_parse = False
        self._uuid = str(uuid.uuid4())
        self._logger = self.startup_logger
        self.status = CanaryRunStatus.NEW

    def _set_logger_meta(self):
        canary_id = self._canary_id
        dag_id = self._canary_dag_id
        run_id = self._canary_run_id
        run_try_number = self._canary_run_try_number

        canary_identifier_fmt = (
            "[canary_id=%s][dag_id=%s][run_id=%s][run_try_number=%s]"
            % (canary_id, dag_id, run_id, run_try_number)
        )

        formatter = logging.Formatter(
            "%(levelname)s %(asctime)s %(name)s.%(funcName)s:%(lineno)s- "
            + canary_identifier_fmt
            + " %(message)s"
        )
        ch = logging.StreamHandler(stream=sys.stdout)
        ch.setFormatter(formatter)
        logger = logging.getLogger(__name__)
        logger.addHandler(ch)
        logging.getLogger("").addHandler(ch)

    @property
    def logger(self):
        if hasattr(self, "_logger"):
            return self._logger
        else:
            return self.startup_logger

    @property
    def _api_key(self):
        return os.getenv("API_KEY")

    @property
    def _target_host(self):
        return os.getenv("TARGET_HOST") or self.target_host

    @property
    def _target_value(self):
        return os.getenv("TARGET_VALUE") or self.target_value

    @property
    def _canary_run_id(self):
        return os.getenv("RUN_ID") or "CANARY_RUN_ID_NOT_FOUND"

    @property
    def _canary_run_try_number(self):
        return os.getenv("TRY_NUMBER") or "CANARY_TRY_NUMBER_NOT_FOUND"

    @property
    def _canary_dag_id(self):
        return os.getenv("DAG_ID") or "CANARY_DAG_ID_NOT_FOUND"

    @property
    def _canaryServer(self):
        """configured via env var"""
        try:
            return os.getenv("CANARY_API_SERVER")
        except Exception as e:
            self.logger.error("no canary server set in env %r", e)

    @property
    def _canary_id(self):
        return os.getenv("CANARY_ID")

    def get_response(self):
        """overwritten by actual canary"""
        raise NotImplementedError

    def parse_response(self, response):
        """overwritten by actual canary"""
        raise NotImplementedError

    def _report(self, error_get=False, error_parse=False, value=None, latency_ms=None):
        """reports results back to canary server"""
        url = os.path.join(self._canaryServer, "canarydata/")

        self._logger.info(
            " _report "
            "[error_get=%s][error_parse=%r][value=%r][url=%s][latency_ms=%s]",
            error_get,
            error_parse,
            value,
            url,
            latency_ms,
        )

        try:
            if hasattr(value, "to_dict"):
                value = value.to_dict()
            _json = {
                "error_get": error_get,
                "error_parse": error_parse,
                "value": value,
                "response_latency": latency_ms,
                "canary_id": self._canary_id,
                "dag_id": self._canary_dag_id,
                "run_id": self._canary_run_id,
                "run_try_number": self._canary_run_try_number,
                "measured": datetime.datetime.utcnow().isoformat(),
                "downtime_generating_error": error_get
            }
            headers = {"Authorization": "Api-Key %s" % self._api_key}

            response = requests.post(url, json=_json, headers=headers)

            self._logger.info(
                "reported [status_code=%s]", response.status_code)

        except Exception as e:
            self._logger.error(
                "exception in _report "
                "[error_get=%s][error_parse=%r][value=%r][latency_ms=%s][error=%s]",
                error_get,
                error_parse,
                value,
                latency_ms,
                e,
            )

    def run(self):
        if IS_AGENT_CANARY and self.first_run:
            self.report_status(CanaryRunStatus.RUNNING)
            self.first_run = False
        self._run()

    def _run(self):
        """gets a response,
        parses it,
        reports parsed results back to server
        """
        value = None
        latency_ms = None
        try:
            self._logger.info("trying get_response")
            start = time.perf_counter()
            response = self.get_response()
            self._logger.info("raw response %s", response)
            self._logger.info("response dir %s", dir(response))

            stop = time.perf_counter()
            latency_ms = round((stop - start) * 1000)
            self._logger.info(
                "get_response [error=False][latency_ms=%s]", latency_ms)
        except Exception as e:
            self._logger.exception("get_response [error=True] %s", e)
            self._error_get = True
        if not self._error_get:
            try:
                self._logger.info("trying parse_response")
                value = self.parse_response(response)
                self._logger.info(
                    "parse_response [error=False][value=%r]", value)
            except Exception as e:
                self._error_parse = True
                self._logger.error("parse_response [error=True] %s", e)

        self._report(self._error_get, self._error_parse, value, latency_ms)

        self.latency_ms = latency_ms
        self.reported_value = value

    def _report_canary_agent_status(self):
        """reports canary agent status to canary server"""
        url = os.path.join(
            self._canaryServer, f"canary/{self._canary_id}/canary_agent_status/"
        )

        self._logger.info(
            f"_report_canary_agent_status [url={url}][status={self.status}]")

        try:
            _json = {"id": self._canary_id, "status": self.status.value}
            headers = {"Authorization": "Api-Key %s" % self._api_key}

            response = requests.patch(url, json=_json, headers=headers)

            self._logger.info(
                "status changed status [status_code=%s]", response.status_code)

        except Exception as e:
            self._logger.error(
                f"exception in _report_canary_agent_status [error={e}]")

    def report_status(self, status):
        if status == CanaryRunStatus.STOPPED:
            self.status = CanaryRunStatus.STOPPED
        elif status == CanaryRunStatus.RUNNING:
            self.status = CanaryRunStatus.RUNNING
        elif status == CanaryRunStatus.ERROR:
            self.status = CanaryRunStatus.ERROR
        elif status == CanaryRunStatus.NEW:
            self.status = CanaryRunStatus.NEW
        else:
            self.status = CanaryRunStatus.NEW

        self._report_canary_agent_status()


""" 
    Below are two concrete examples of a Canary class
"""


class HTTPGetCanary(CanaryBasePrototype):
    """runs an HTTP get and returns response"""

    def get_response(self):
        return requests.get(self._target_host)


class HTTPGetResponseBodyTextCanary(HTTPGetCanary):
    """pings an nginx instance test url and tests for text in response body"""

    def parse_response(self, response):
        self._logger.info(
            "parse_response [target_value=%s]", self._target_value)
        self._logger.info(f"parse_response [response.text={response.text}]")
        try:
            text = response.text
        except AttributeError:
            raise AttributeError("No Response Text")

        return text

class HTTPGetResponseBodyTextBooleanCanary(HTTPGetCanary):
    """pings an nginx instance test url and tests for text in response body"""

    def parse_response(self, response):
        self._logger.info(
            "parse_response [target_value=%s]", self._target_value)
        self._logger.debug("content %s", response.text)
        try:
            target_value = os.getenv("TARGET_VALUE", self._target_value)
        except AttributeError:
            raise AttributeError("No Target Value Set")
        try:
            text = response.text
        except AttributeError:
            raise AttributeError("No Response Text")

        return bool(target_value in text)


class HTTPGetStatusCodeCanary(HTTPGetCanary):
    """Checks status code from an HTTP Response"""

    def parse_response(self, response):
        return response.status_code


if __name__ == "__main__":

    if CANARY_TYPE_NAME == "HTTP_RESPONSE_STATUS_CODE":
        canary = HTTPGetStatusCodeCanary

    elif CANARY_TYPE_NAME == "HTTP_RESPONSE_BODY":
        canary = HTTPGetResponseBodyTextCanary

    elif CANARY_TYPE_NAME == "HTTP_RESPONSE_BODY_BOOLEAN":
        canary = HTTPGetResponseBodyTextBooleanCanary

    if IS_AGENT_CANARY:
        killer = GracefulKiller()
        nextRunTime = getNextCronRunTime(CANARY_SCHEDULE)
        while not killer.kill_now:
            roundedDownTime = roundDownTime()
            if roundedDownTime == nextRunTime:
                canary().run()
                nextRunTime = getNextCronRunTime(CANARY_SCHEDULE)
            elif roundedDownTime > nextRunTime:
                nextRunTime = getNextCronRunTime(CANARY_SCHEDULE)
            sleepTillTopOfNextMinute()
        canary().report_status(CanaryRunStatus.STOPPED)
    else:
        canary().run()
